<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.1/2030377
 */

use Api\Controller\IndexController as ApiController;
use Api\IRequest;
use Application\Config\ConfigManager;
use Application\Model\IModelDAO;
use Reviews\Model\ReviewDAO;
use Application\Controller\IndexControllerFactory;
use Application\Factory\InvokableServiceFactory;
use Application\View\Helper\ViewHelperFactory;
use Application\Config\Services;
use Reviews\Controller\ReviewApi;
use Reviews\Controller\IndexController;
use Reviews\Filter\Keywords;
use Reviews\Filter\Participants;
use Reviews\Filter\IParticipants;
use Reviews\Validator\Transitions;
use Reviews\Filter\VoteInput;
use Laminas\Http\Request;
use Reviews\Filter\IVersion;
use Reviews\Filter\Version;

return [
    ConfigManager::REVIEWS => [
        'patterns' => [
            'octothorpe' => [     // #review or #review-1234 with surrounding whitespace/eol
                'regex'  => '/(?P<pre>(?:\s|^)\(?)'
                    . '\#(?P<keyword>review|append|replace)(?:-(?P<id>[0-9]+))?'
                    . '(?P<post>[.,!?:;)]*(?=\s|$))/i',
                'spec'   => '%pre%#%keyword%-%id%%post%',
                'insert' => "%description%\n\n#review-%id%",
                // Ignore line limit here to aid readability
                // @codingStandardsIgnoreStart
                'strip'  => '/^\s*\#(review|append|replace)(-[0-9]+)?(\s+|$)|(\s+|^)\#(review|append|replace)(-[0-9]+)?\s*$/i'
                // @codingStandardsIgnoreEnd
            ],
            'leading-square'  => [ // [review] or [review-1234] at start
                'regex'  => '/^(?P<pre>\s*)\[(?P<keyword>review|append|replace)(?:-(?P<id>[0-9]+))?\](?P<post>\s*)/i',
                'spec'   => '%pre%[%keyword%-%id%]%post%'
            ],
            'trailing-square' => [ // [review] or [review-1234] at end
                'regex'  => '/(?P<pre>\s*)\[(?P<keyword>review|append|replace)(?:-(?P<id>[0-9]+))?\](?P<post>\s*)?$/i',
                'spec'   => '%pre%[%keyword%-%id%]%post%'
            ],
        ],
        'filters' => [
            'fetch-max' => 50,
            'filter-max' => 50,
            'result_sorting' => true,
            'date_field' => 'created', // 'created' displays and sorts by created date, 'updated' displays and sorts
            // by last updated
            // These need to match Review::FETCH_BY...
            'hasVoted' => [
                'fetch-max' => 5000,
            ],
            'myComments' => [
                'fetch-max' => 2500,
            ],
        ],
        'expand_group_reviewers' => false, // whether swarm should expand group members on the review page if they
        // have been added as an individual.
        'cleanup'              => [
            'mode'        => 'user', // auto - follow default, user - present checkbox(with default)
            'default'     => false,  // clean up pending changelists on commit
            'reopenFiles' => false   // re-open any opened files into the default changelist
        ],
        'disable_commit'        => false,
        'disable_self_approve'  => false, // whether authors can approve their own reviews
        'commit_credit_author'  => true,
        'commit_timeout'        => 1800,  // default: 30 minutes (must be in seconds)
        'unapprove_modified'    => true,  // whether approved reviews with modified files can be automatically
        // unapproved
        'ignored_users'         => [],
        'allow_author_change'   => false, // Whether anyone can change the Author
        'sync_descriptions'     => false, // if true a changesaved event will update all reviews attached to said change
        'expand_all_file_limit' => 10,    // Controls if 'Expand all' is available for reviews by specifying a file
        // limit over which the option will not be available. 0 signifies always on.
        'process_shelf_delete_when' => [], // States of the review we will process if a user deletes files from
        // their shelved changelist.
        // Supported states array('needsReview', 'needsRevision', 'archived', 'rejected', 'approved')
        'disable_approve_when_tasks_open' => false, // false shows a warning if tasks are open, true prevents approve
        'version_chooser'                 => 'chooser',
        'more_context_lines'              => 10,
        'max_bottom_context_lines'        => 10000,
        'allow_author_obliterate'         => false, // If true author will be allowed to obliterate their own review.
        'moderator_approval'              => ConfigManager::VALUE_ANY, // for reviews with multiple branches either
        // 'any' moderator may approve or 'each' must
        // approve (at least one from each branch)
        'end_states'                      => [],  // States that are considered closed
        // when checking workflow rules to
        // see if a review can be changed
        ConfigManager::REACT_ENABLED      => false,  // Prefer php based reviews
        ConfigManager::ALLOW_EDITS        => true, // Allow inline editing of files in a review
    ],
    'security' => [
        'login_exempt'  => ['review-tests', 'review-deploy']
    ],
    'router' => [
        'routes' => [
            'api' => [
                'type' => 'literal',
                'options' => [
                    'route' => ApiController::API_BASE,
                ],
                'may_terminate' => false,
                'child_routes' => [
                    'reviews' => [
                        'type' => 'Laminas\Router\Http\Segment',
                        'options' => [
                            'route' => '/:version/reviews/:id',
                            'constraints' => [IRequest::VERSION => 'v10'],
                            'defaults' => [
                                'controller' => ReviewApi::class
                            ],
                        ],
                        'child_routes' => [
                            'review' => [
                                'type' => 'Laminas\Router\Http\Method',
                                'options' => [
                                    'verb' => 'get',
                                ]
                            ],
                            'vote' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/vote'
                                ],
                                'child_routes' => [
                                    'vote-up-down' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'post',
                                            'defaults' => [
                                                'action' => 'vote'
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                            'review-author' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/author'
                                ],
                                'child_routes' => [
                                    'review-author-update' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'put',
                                            'defaults' => [
                                                'action' => 'author'
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                            'delete-review' => [
                                'type' => 'Laminas\Router\Http\Method',
                                'options' => [
                                    'verb' => Request::METHOD_DELETE,
                                    'defaults' => [
                                        'controller' => ReviewApi::class
                                    ],
                                ],
                            ],
                            'transitions' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/transitions'
                                ],
                                'child_routes' => [
                                    'transitions' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'get',
                                            'defaults' => [
                                                'action' => 'transitions'
                                            ],
                                        ],
                                    ]
                                ],
                            ],
                            'transition' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/transition'
                                ],
                                'child_routes' => [
                                    'transition' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'post',
                                            'defaults' => [
                                                'action' => 'transition'
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                            'review-description' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/description'
                                ],
                                'child_routes' => [
                                    'update' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'put',
                                            'defaults' => [
                                                'action' => 'description'
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                            'participants' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/participants',
                                ],
                                'child_routes' => [
                                    'add' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => Request::METHOD_POST,
                                            'defaults' => [
                                                'action' => 'addParticipants'
                                            ],
                                        ],
                                    ],
                                    'update' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => Request::METHOD_PUT,
                                            'defaults' => [
                                                'action' => 'updateParticipants'
                                            ],
                                        ],
                                    ],
                                    'delete' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => Request::METHOD_DELETE,
                                            'defaults' => [
                                                'action' => 'deleteParticipants'
                                            ],
                                        ],
                                    ],
                                ],
                            ],
                            'archive' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/archive'
                                ],
                                'child_routes' => [
                                    'transitions' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'get',
                                            'defaults' => [
                                                'action' => 'archive'
                                            ],
                                        ],
                                    ]
                                ],
                            ],
                            'files' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/files'
                                ],
                                'child_routes' => [
                                    'fileChanges' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'get',
                                            'defaults' => [
                                                'action' => 'fileChanges'
                                            ],
                                        ],
                                    ],
                                ]
                            ],
                            'comments' => [
                                'type' => 'Laminas\Router\Http\Segment',
                                'options' => [
                                    'route' => '/comments'
                                ],
                                'child_routes' => [
                                    'transitions' => [
                                        'type' => 'Laminas\Router\Http\Method',
                                        'options' => [
                                            'verb' => 'get',
                                            'defaults' => [
                                                'action' => 'getComments'
                                            ],
                                        ],
                                    ]
                                ],
                            ],
                        ],
                    ],
                ]
            ],
            'review' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review[/v[:version]][/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'review',
                        'version'    => null
                    ],
                ],
            ],
            'review-history ' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review[/v[:version]]/history[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'review',
                        'version'    => null
                    ],
                ],
            ],
            'review-files ' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review[/v[:version]]/files[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'review',
                        'version'    => null
                    ],
                ],
            ],
            'review-comments ' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review[/v[:version]]/comments[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'review',
                        'version'    => null
                    ],
                ],
            ],
            'review-version-delete' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/v:version/delete[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'deleteVersion',
                        'version'    => null
                    ],
                ],
            ],
            'review-reviewer' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/reviewers/:user[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'reviewer',
                        'user'       => null
                    ],
                ],
            ],
            'review-reviewers' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/reviewers[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'reviewers'
                    ],
                ],
            ],
            'review-author' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/author[/:author][/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'editAuthor',
                        'author'     => null
                    ],
                ],
            ],
            'review-vote' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/vote/:vote[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'vote',
                        'vote'       => null
                    ],
                ],
            ],
            'review-tests' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/tests/:status[/:token][/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'testStatus',
                        'status'     => null,
                        'token'      => null
                    ],
                ],
            ],
            'review-deploy' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/deploy/:status[/:token][/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'deployStatus',
                        'status'     => null,
                        'token'      => null
                    ],
                ],
            ],
            'review-transition' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/:review/transition[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'transition'
                    ],
                ],
            ],
            'dashboards' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route' => '/dashboards/action',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'dashboard',
                        'author' => null
                    ],
                ],
            ],
            'reviews' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s][/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'index'
                    ],
                ],
            ],
            'add-review' => [
                'type' => 'Laminas\Router\Http\Segment',
                'options' => [
                    'route'    => '/review[s]/add[/]',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'add'
                    ],
                ],
            ],
            'review-file' => [
                'type' => 'Application\Router\Regex',
                'options' => [
                    'regex'    => '/reviews?/(?P<review>[0-9]+)/v(?P<version>[0-9,]+)/files?(/(?P<file>.*))?',
                    'spec'     => '/reviews/%review%/v%version%/files/%file%',
                    'defaults' => [
                        'controller' => IndexController::class,
                        'action'     => 'fileInfo',
                        'review'     => null,
                        'version'    => null,
                        'file'       => null
                    ],
                ],
            ],
        ],
    ],
    'controllers' => [
        'factories' => [
            IndexController::class => IndexControllerFactory::class,
            ReviewApi::class => IndexControllerFactory::class,
        ],
    ],
    'service_manager' => [
        'aliases' => [
            Keywords::SERVICE     => Keywords::class,
            Services::TRANSITIONS => Transitions::class,
            IModelDAO::REVIEW_DAO => ReviewDAO::class,
            Services::VOTE_INPUT_FILTER => VoteInput::class,
            IParticipants::PARTICIPANTS => Participants::class,
            IVersion::VERSION_FILTER => Version::class
        ],
        'factories' => array_fill_keys(
            [
                Keywords::class,
                Transitions::class,
                ReviewDAO::class,
                VoteInput::class,
                Participants::class,
                Version::class
            ],
            InvokableServiceFactory::class
        )
    ],
    'view_manager' => [
        'template_path_stack'   => [
            __DIR__ . '/../view',
        ],
    ],
    'view_helpers' => [
        'factories' => array_fill_keys(
            [
                ViewHelperFactory::REVIEWS,
                ViewHelperFactory::REVIEW_KEYWORDS,
                ViewHelperFactory::REVIEWERS_CHANGES,
                ViewHelperFactory::AUTHOR_CHANGE,
            ],
            ViewHelperFactory::class
        )
    ],
    'menu_helpers' => [
        'reviews' => [
            'class'    => '\Projects\Menu\Helper\ProjectAwareMenuHelper',
            'priority' => 130
        ]
    ]
];
