<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2020.1/2030377
 */
namespace Search\Controller;

use Api\Controller\AbstractRestfulController;
use Application\Config\ConfigManager;
use Application\Config\IConfigDefinition;
use Application\Config\IDao;
use Application\Helper\BooleanHelper;
use Application\I18n\TranslatorFactory;
use Redis\Model\AbstractDAO;
use Laminas\Http\Response;

class SearchApi extends AbstractRestfulController
{
    // API Request Params
    const API_VERSION      = "v10";
    const SEARCH_TERM      = 'term';
    const SEARCH_CONTEXT   = 'context';
    const LIMIT            = 'limit';
    const STARTS_WITH_ONLY = 'starts_with_only';
    const IGNORE_BLACKLIST = 'ignore_blacklist';

    // Error code
    const ERROR_CODE = 'search-error';

    // Contexts that support searching
    const CONTEXT_MAP = [
        'user'  => [
            'dao'       => IDao::USER_DAO,
            'blacklist' => ConfigManager::MENTIONS_USERS_BLACKLIST
        ],
        'group' => [
            'dao' => IDao::GROUP_DAO,
            'blacklist' => ConfigManager::MENTIONS_GROUPS_BLACKLIST
        ],
    ];

    /**
     * This API is used to search over a specified context and return records matching the given search term.
     *
     * It can accept two types of search:
     *      1) Includes:           A record is considered matched if it includes the search term anywhere within it.
     *                             This is the default and will actually run a starts with search first.
     *      2) Starts with only:   A record is only considered matched if it begins with the search term. This is a much
     *                             faster search. This type of search can be specified with the start_with_only param
     *
     * Supported context models are currently only user and group.
     *
     * Additionally, the total number of records returned can be limited by specifying a limit. Ten is the default.
     * If the number of matches is more than the limit, matches are truncated from the last context to the first.
     *
     * Furthermore, unless an 'ignore_blacklist' param is specified with a value of true, any corresponding blacklists
     * will be applied.
     *
     * The searches are case-insensitive.
     *
     * Matches are returned as follows:
     *      For user:  matches are returned as arrays of usernames and full names
     *      For group: matches are returned as arrays of groupIds and group names
     *      If a limit is specified in the request, it is also returned with the data
     *
     * Examples:
     *      Request:  /api/v10/search?starts_with_only=true&term=J&context=user,group&limit=3
     *      Response: [
     *                    'data' => [
     *                        'user'  => [
     *                            [
     *                                'User'     => 'jimbob',
     *                                'FullName' => 'Frank Smith'
     *                            ],
     *                            [
     *                                'User'     => 'J1Mb0b77',
     *                                'FullName' => 'Jimbob ANDREWS'
     *                            ]
     *                        ],
     *                        'group' => [
     *                            [
     *                                'Group' =>'jjp',
     *                                'name'  =>'JJPs Group',
     *                            ]
     *                        ],
     *                        'limit'  => 3
     *                    ]
     *                ]
     *      Request:  /api/v10/search?&term=J&context=user,group&limit=3
     *      Response: [
     *                    'data' => [
     *                        'user'  => [
     *                            [
     *                                'User'     => 'jimbob',
     *                                'FullName' => 'Frank Smith'
     *                            ],
     *                            [
     *                                'User'     => 'raj',
     *                                'FullName' => 'Raja Paresh'
     *                            ]
     *                        ],
     *                        'group' => [
     *                            [
     *                                'Group' =>'sojourners',
     *                                'name'  =>'Sojourners',
     *                            ]
     *                        ],
     *                        'limit'  => 3
     *                    ]
     *                ]
     */
    public function searchAction()
    {
        // Returned with success response
        $data = [];

        // Request params
        $request         = $this->getRequest();
        $term            = $request->getQuery(self::SEARCH_TERM);
        $originalContext = $request->getQuery(self::SEARCH_CONTEXT);
        $context         = $originalContext ? explode(',', $originalContext) : [];
        $limit           = $request->getQuery(self::LIMIT, '0');
        $limit           = $limit === '' ? '0' : $limit;
        $startsWithOnly  = BooleanHelper::isTrue($request->getQuery(self::STARTS_WITH_ONLY));
        $ignoreBlacklist = BooleanHelper::isTrue($request->getQuery(self::IGNORE_BLACKLIST));

        // Validate the request
        $errors = $this->validateRequest($term, $limit, $originalContext, $context);
        if (isset($errors)) {
            return self::error($errors, Response::STATUS_CODE_422);
        }

        // Config for use in getting blacklists
        $config = $this->services->get(IConfigDefinition::CONFIG);

        // Get matches for each model
        foreach ($context as $model) {
            if (isset(self::CONTEXT_MAP[$model])) {
                $data[$model] = $this->recordsForModel($model, $term, $startsWithOnly, $config, $ignoreBlacklist);
            }
        }

        // Now we impose the limit on the overall results
        $this->limitResults($data, $limit);
        return self::success($data);
    }

    /**
     * Validates the request and builds messages when not valid
     * @param mixed   $term              term that we are searching for
     * @param mixed   $limit             maximum number of results to return
     * @param mixed   $originalContext   context over which we are searching, in string form
     * @param mixed   $context           context over which we are searching, in array form
     * @return array|null
     */
    protected function validateRequest($term, $limit, $originalContext, $context)
    {
        $messages   = null;
        $translator = $this->services->get(TranslatorFactory::SERVICE);

        if (is_null($term) || $term === '') {
            $messages[] = $this->buildMessage(self::ERROR_CODE, $translator->t('Must include a search term'));
        }

        $validContext  = array_keys(self::CONTEXT_MAP);
        $invalidModels = array_diff($context, $validContext);

        if (count($context) === 0 || count($invalidModels) > 0) {
            $validString = implode(', ', $validContext);
            $msg         = "Invalid context [$originalContext], must contain only [$validString], separated by commas";
            $messages[]  = $this->buildMessage(self::ERROR_CODE, $translator->t($msg));
        }

        // Since $limit is coming from the request, we are guaranteed a string
        if (!ctype_digit($limit) || (int)$limit < 0) {
            $msg        = "Invalid limit [$limit], must be a non-negative integer";
            $messages[] = $this->buildMessage(self::ERROR_CODE, $translator->t($msg));
        }

        return $messages;
    }

    /**
     * Calls the fetchAll method of the given DAO model and passes it the appropriate params
     *
     * @param string   $model             model over whose records we are searching
     * @param string   $term              term that we are searching for
     * @param bool     $startsWithOnly    optionally, specify if you only want to search for records starting with the
     *                                    given term
     * @param array    $config            configuration array
     * @param bool     $ignoreBlacklist   whether blacklists should be ignored
     *
     * @throws \Application\Config\ConfigException
     * @return array
     */
    protected function recordsForModel($model, $term, $startsWithOnly, $config, $ignoreBlacklist)
    {
        $dao       = $this->services->get(self::CONTEXT_MAP[$model]['dao']);
        $blacklist = $ignoreBlacklist
            ? []
            : ConfigManager::getValue($config, self::CONTEXT_MAP[$model]['blacklist'], []);

        $options = [
            AbstractDAO::FETCH_SEARCH => [
                AbstractDAO::SEARCH_TERM               => $term,
                AbstractDAO::SEARCH_LIMIT              => 0,
                AbstractDAO::SEARCH_RETURN_RAW_ENTRIES => false,
                AbstractDAO::SEARCH_STARTS_WITH_ONLY   => $startsWithOnly,
                AbstractDAO::SEARCH_BLACKLIST          => $blacklist
            ]
        ];

        return $dao->fetchAll($options);
    }

    /**
     * Limit the results according to the order in which the the models were listed in the context
     * If a limit is specified, distribute the results as evenly as possible across the models and
     * add it to the data array as a key-value pair.
     *
     * @param array   $data    data to be ordered
     * @param int     $limit   total number of results returned
     */
    protected function limitResults(array &$data, int $limit)
    {
        // If there is not a positive limit, we don't limit results and this is a noop
        if ($limit <= 0) {
            return;
        }

        // Create some objects to keep track of state during the loop
        $metrics = [];
        foreach ($data as $model => $matches) {
            $metrics[$model] = ['numMatches' => count($matches), 'limit' => 0, 'done' => false];
        }

        // The amount of remaining limit to allocate
        $remainingLimit = $limit;
        // The number of models having more records than their allotted limit
        $numNeedingMoreLimit = count($metrics);

        // iteratively allocate the remaining limit until there's no more remaining limit or all models have allocations
        while ($remainingLimit > 0 && $numNeedingMoreLimit > 0) {
            $allocations = array_reverse($this->getAllocations($remainingLimit, $numNeedingMoreLimit));
            foreach ($metrics as $model => $metric) {
                if ($metric['done'] === false) {
                    $allocation       = array_pop($allocations);
                    $metric['limit'] += $allocation;
                    if ($metric['numMatches'] <= $metric['limit']) {
                        // The model has all its matches covered by its limit allotment
                        // and will only reduce the remaining limit by its number of matchhes
                        $remainingLimit -= $metric['numMatches'];
                        --$numNeedingMoreLimit;
                        $metric['done'] = true;
                    } else {
                        // The model will use all the allotment and reduce the remaining limit by that much
                        $remainingLimit -= $allocation;
                    }
                }
                $metrics[$model] = $metric;
            }
        }

        // Apply the limits
        foreach ($data as $model => $matches) {
            $data[$model] = array_slice($data[$model], 0, $metrics[$model]['limit']);
        }

        // Add the limit to the data returned
        $data[self::LIMIT] = $limit;
    }

    /**
     * Distributes an integer into an array of integer addends as equally as possible.
     * This algorithm divides the numerator by the denominator and if there is a remainder,
     * distributes it from left to right.
     *
     * @param int   $numerator     number to be divided
     * @param int   $denominator   number of elements in the resulting array
     *
     * @return array
     */
    protected function getAllocations($numerator, $denominator)
    {
        $allocations = [];
        $intDiv      = intdiv($numerator, $denominator);
        $rem         = $numerator % $denominator;
        for ($i = 0; $i < $denominator; $i++) {
            $allocations[$i] = $intDiv;
        }
        for ($i = 0; $i < $rem; $i++) {
            ++$allocations[$i];
        }
        return $allocations;
    }
}
